# coding: utf-8
module Gomasio
  #
  class Game
    attr_reader :order
    
    #
    Command = Struct.new(:row, :col, :disk)
    
    #
    Status = Struct.new(:retry, :in_same, :out_range) do
      def clear
        @retry     = false
        @in_same   = false
        @out_range = false
      end
    end

    #
    def initialize(size: 19, order: Gomasio::Disk.black)
      @board = Gomasio::Board.new(size)
      @order = order

      yield self if block_given?

      @runloop = Fiber.new do
        @board.clear
        runloop
      end
    end

    #
    def <<(other)
      fail ArgumentError.new unless other[:player].is_a? Gomasio::Player
      fail ArgumentError.new unless other[:order].is_a? Gomasio::Disk

      @players ||= {}
      @players[other[:order]] = other[:player]
    end

    #
    def size
      @board.size
    end

    #
    def ready?
      @players.count == 2
    end

    def board
      @board.to_matrix
    end

    #
    def run
      @runloop.resume
    end

    private

    #
    def runloop
      @commands ||= []
      status = Status.new
      begin
        loop do
          pos = @players[@order].turn_responding(order: @order, status: status, board: @board.clone)
          raise OutOfRangeError.new unless @board.include?(pos.row, pos.col)

          _, win = @board.mount(row: pos.row, col: pos.col, disk: @order)
          @commands << Command.new(pos.row, pos.col, @order)
          status.clear

          Fiber.yield(win)
          @order = @order.reverse
        end
      rescue InSamePlaceError, OutOfRangeError => e
        status.retry     = true
        status.in_same   = true if e.is_a? InSamePlaceError
        status.out_range = true if e.is_a? OutOfRangeError
        retry
      end
    end
  end
end
